/*
восприятие действий и фраз с Пусльта

Создание образов различной иерархии контекстов восприятия:
BaseStyleArr - образ сочетаний активных Базовых контекстов
TriggerStimulsArr - образ сочетаний пусковых стимулов

Создание образов сочетаний ID действий:
*/

package reflexes

import (
	"BOT/brain/action_sensor"
	 "BOT/brain/gomeostas"
	wordSensor "BOT/brain/words_sensor"
	"BOT/lib"
	"strconv"
	"strings"
)

///////////////////////////////////////
func loadImagesArrs(){
	//  загрузить образы сочетаний базовых стилей
	loadBaseStyleArr()
	//  загрузить образы сочетаний пусковых стимулов
	loadTriggerStimulsArr()

}
/////////////////////////////////////////////
// ПУЛЬС рефлексов
var ReflexPulsCount=0 // передача тика Пульса из brine.go
// коррекция текущего состояния гомеостаза и базового контекста с каждым пульсом
func ReflexCountPuls(puls int){
	ReflexPulsCount=puls // передача номера тика из более низкоуровневого пакета

	if gomeostas.IsNewConditions{// изменились условия
		ActiveFromConditionChange()
		return
	}

	// обнулить причину возможного запуска рефлекса
	if oldActiveCurTriggerStimulsID>0 && oldActiveCurTriggerStimulsPulsCount>(ReflexPulsCount+10){
		oldActiveCurTriggerStimulsID=0
	}
}
//////////////////////////////////////////////////////////////



/////////////////  сохранить образы сочетаний базовых стилей
func SaveBaseStyleArr(){
	var out=""
	for k, v := range BaseStyleArr {
		out+=strconv.Itoa(k)+"|"
		for i := 0; i < len(v.BSarr); i++ {
			out+=strconv.Itoa(v.BSarr[i])+","
		}
		out+="\r\n"
	}
	lib.WriteFileContent(lib.GetMainPathExeFile()+"/memory_reflex/base_style_images.txt",out)
}
//////////////////  загрузить образы сочетаний базовых стилей
func loadBaseStyleArr(){
	BaseStyleArr=make(map[int]*BaseStyle)
	strArr,_:=lib.ReadLines(lib.GetMainPathExeFile()+"/memory_reflex/base_style_images.txt")
	cunt:=len(strArr)
	for n := 0; n < cunt; n++ {
		if len(strArr[n])==0{
			continue
		}
		p:=strings.Split(strArr[n], "|")
		id,_:=strconv.Atoi(p[0])
		s:=strings.Split(p[1], ",")
		var BSarr[] int
		for i := 0; i < len(s); i++ {
			if len(s[i])==0{
				continue
			}
			si,_:=strconv.Atoi(s[i])
			BSarr=append(BSarr,si)
		}
		createNewBaseStyle(id,BSarr)
	}
	return
}
/////////////////////////////////////////////////////////////////////




/////////////////////////////////////////
// базовый стиль - Образ сочетаний Базовых Контекстов гомеостаза
type BaseStyle struct {
	ID int  // идентификатор данного сочетания контекстов
	BSarr[] int
}
var BaseStyleArr=make(map[int]*BaseStyle)

// создать образ сочетаний активных Базовых контекстов
var lastBaseStyleID=0
func createNewBaseStyle(id int,BSarr []int)(int,*BaseStyle){
	oldID,oldVal:=checkUnicumBaseStyle(BSarr)
	if oldVal!=nil{
		return oldID,oldVal
	}
	if id==0{
		lastBaseStyleID++
		id=lastBaseStyleID
	}else{
		//		newW.ID=id
		if lastBaseStyleID<id{
			lastBaseStyleID=id
		}
	}

	var node BaseStyle
	node.ID = id
	node.BSarr = BSarr

	BaseStyleArr[id]=&node
	return id,&node
}
func checkUnicumBaseStyle(bArr []int)(int,*BaseStyle){
	for id, v := range BaseStyleArr {
		if lib.EqualArrs(bArr,v.BSarr) {
			return id,v
		}
	}
return 0,nil
}
////////////////////////////////////////


/* Образ текущего сочетания пусковых (Trigger) стимулов в восприятии
Сочетание пусковых стимулов, включая фразы и включая тон, настроение - должны быть уникальными
чтобы только для данного сочетания мог срабатывать данный рефлекс.

Это - образ много-модеального воздействия или просто образ восприятия.
 */
type TriggerStimuls struct {
	ID int  // идентификатор данного сочетания пусковых стимулов
	RSarr[] int // массив действий с Пульта
	// для текущей фразы:
	PhraseID []int
	ToneID int
	MoodID int
}
var TriggerStimulsArr=make(map[int]*TriggerStimuls)
//////////////////////////////////////////

// создать образ сочетаний пусковых стимулов
//В случае отсуствия пусковых стимулов создается ID такого отсутсвия, пример такой записи: 2|||0|0| - ID=2
var lastTriggerStimulsID=0
func createNewlastTriggerStimulsID(id int,RSarr []int,PhraseID []int,ToneID int,MoodID int)(int,*TriggerStimuls){
	oldID,oldVal:=checkUnicumTriggerStimuls(RSarr,PhraseID,ToneID,MoodID)
	if oldVal!=nil{
		return oldID,oldVal
	}
	if id==0{
		lastTriggerStimulsID++
		id=lastTriggerStimulsID
	}else{
		//		newW.ID=id
		if lastTriggerStimulsID<id{
			lastTriggerStimulsID=id
		}
	}

	var node TriggerStimuls
	node.ID = id
	node.RSarr = RSarr
	node.PhraseID=PhraseID
	node.ToneID=ToneID
	node.MoodID=MoodID

	TriggerStimulsArr[id]=&node
	return id,&node
}
func checkUnicumTriggerStimuls(bArr []int,PhraseID []int,ToneID int,MoodID int)(int,*TriggerStimuls){
	for id, v := range TriggerStimulsArr {
		if !lib.EqualArrs(bArr,v.RSarr) {
			continue
		}
		if !lib.EqualArrs(PhraseID,v.PhraseID) {
			continue
		}
		if ToneID!=v.ToneID || MoodID!=v.MoodID {
			continue
		}
		return id,v
	}

	return 0,nil
}
/////////////////////////////////////////
// создать новое сочетание пусковых стимулов если такого еще нет
func CreateNewTriggerStimulsImage(){
	PhraseID:=wordSensor.CurrentPhrasesIDarr
	ToneID:=wordSensor.DetectedTone
	MoodID:=wordSensor.CurPultMood
	RSarr:=action_sensor.CheckCurActions()

	setOldActiveCurTriggerStimulsVal(ActiveCurTriggerStimulsID)
	ActiveCurTriggerStimulsID,_=createNewlastTriggerStimulsID(0,RSarr,PhraseID,ToneID,MoodID)

	saveTriggerStimulsArr()
}

/////////////////////////////////////////





//////////////////// сохранить образы сочетаний пусковых стимулов
//В случае отсуствия пусковых стимулов создается ID такого отсутсвия, пример такой записи: 2|||0|0|
func saveTriggerStimulsArr(){
	var out=""
	for k, v := range TriggerStimulsArr {
		out+=strconv.Itoa(k)+"|"
		for i := 0; i < len(v.RSarr); i++ {
			out+=strconv.Itoa(v.RSarr[i])+","
		}
		out+="|"
		for i := 0; i < len(v.PhraseID); i++ {
			out+=strconv.Itoa(v.PhraseID[i])+","
		}
		out+="|"
		out+=strconv.Itoa(v.ToneID)+"|"
		out+=strconv.Itoa(v.MoodID)+"|"
		out+="\r\n"
	}
	lib.WriteFileContent(lib.GetMainPathExeFile()+"/memory_reflex/trigger_stimuls_images.txt",out)

}
////////////////////  загрузить образы сочетаний пусковых стимулов
func loadTriggerStimulsArr(){
	TriggerStimulsArr=make(map[int]*TriggerStimuls)
	strArr,_:=lib.ReadLines(lib.GetMainPathExeFile()+"/memory_reflex/trigger_stimuls_images.txt")
	cunt:=len(strArr)
	for n := 0; n < cunt; n++ {
		if len(strArr[n])==0{
			continue
		}
		p:=strings.Split(strArr[n], "|")
		id,_:=strconv.Atoi(p[0])

		s:=strings.Split(p[1], ",")
		var RSarr[] int
		for i := 0; i < len(s); i++ {
			if len(s[i])==0{
				continue
			}
			si,_:=strconv.Atoi(s[i])
			RSarr=append(RSarr,si)
		}

		s=strings.Split(p[2], ",")
		var PhraseID[] int
		for i := 0; i < len(s); i++ {
			if len(s[i])==0{
				continue
			}
			si,_:=strconv.Atoi(s[i])
			PhraseID=append(PhraseID,si)
		}
		x,_:=strconv.Atoi(p[3])
		ToneID:=x
		x,_=strconv.Atoi(p[4])
		MoodID:=x

		createNewlastTriggerStimulsID(id,RSarr,PhraseID,ToneID,MoodID)
	}
	return

}
///////////////////////////////////////////





//////  АКТИВАЦИЯ ДЕРЕВА РЕФЛЕКСОВ ПО изменению условий, действиям с Пульта или фразе с Пульта

/*  Вид активации дерева рефлексов:
1 - изменение сочетания базовых контекстов
2 - действия с Пульта
3 - фраза с Пульта
 */
var ActivationTypeSensor=0


////////////////////////////////////
// текущее восприятие ID образов
//обновляющихся при каждом событии с Пульта или достаточно сильном изменении Базовых параметров
var ActiveCurBaseID=0 // ID Базового состояния CommonBadNormalWell
var ActiveCuraseStyleID=0 // ID сочетания базовых контекстов BaseStyle
var ActiveCurTriggerStimulsID=0 // ID теущего активного образа сочетаний пусковых стимулов TriggerStimuls
/* предыдущий образ сочетания пусковых стимулов
используется как ПРИЧИНА последующих событий - для формирования условных рефлексов
потому как ОБНУЛЯЕТСЯ при:
1) активации дерева рефлексов, если вызвало какое-то действие
2) через 10 пульсов после записи значения - типа причина устаревает
 */
var oldActiveCurTriggerStimulsID=0
func setOldActiveCurTriggerStimulsVal(val int){
	oldActiveCurTriggerStimulsID=val
	oldActiveCurTriggerStimulsPulsCount=ReflexPulsCount
}
// момент записи значения в тике Пульса
var oldActiveCurTriggerStimulsPulsCount=0

////////////////////////
/// Дерево рефлексов активируется при любом изменении условий с проверкой по каждому пульсу:
func ActiveFromConditionChange(){
	ActiveCurBaseID=gomeostas.CommonBadNormalWell
	// определение текущего сочетания ID Базовых контекстов
	bsIDarr:=gomeostas.GetCurContextActiveIDarr()
	// создаем новый образ Базовых контекстов, если такого еще нет
	ActiveCuraseStyleID,_=createNewBaseStyle(0,bsIDarr)

	// активировать дерево
	ActivationTypeSensor=1
	activeReflexTree()
}
////////////////////////
// активировать дерево действием reflexes.ActiveFromAction()
func ActiveFromAction(){
	ActiveCurBaseID=gomeostas.CommonBadNormalWell
	// определение текущего сочетания ID Базовых контекстов
	bsIDarr:=gomeostas.GetCurContextActiveIDarr()
	// создаем новый образ Базовых контекстов, если такого еще нет
	ActiveCuraseStyleID,_=createNewBaseStyle(0,bsIDarr)

	// создаем новый образ Базовых контекстов, если такого еще нет
	CreateNewTriggerStimulsImage()

	// активировать дерево
	ActivationTypeSensor=2
	activeReflexTree()

}
////////////////////////
// активировать дерево фразой  reflexes.ActiveFromPhrase() - только для условных рефлексов
func ActiveFromPhrase(){
	ActiveCurBaseID=gomeostas.CommonBadNormalWell
	// определение текущего сочетания ID Базовых контекстов
	bsIDarr:=gomeostas.GetCurContextActiveIDarr()
	// создаем новый образ Базовых контекстов, если такого еще нет
	ActiveCuraseStyleID,_=createNewBaseStyle(0,bsIDarr)
	// создать новое сочетание пусковых стимулов если такого еще нет
	CreateNewTriggerStimulsImage()

	// активировать дерево
	//BS:=gomeostas.GetCurContextActiveIDarr() // текущий стиль Бозовых контекстов
	ActivationTypeSensor=3
	activeReflexTree()
}
/////////////////////////////////////////////////////////




//////////////////////////////////////////////////////////////
/* создание иерархии образов контекстов условий и пусковых стимулов в виде ID образов в [3]int
 создать последовательность уровней условий в виде массива  ID последовательности ID уровней
В случае отсуствия пусковых стимулов создается ID такого отсутсвия, пример такой записи: 2|||0|0|
 */
func getConditionsArr(lev1ID int, lev2 []int, lev3 []int, PhraseID []int,ToneID int,MoodID int )([]int){
	arr:=make([]int, 3)
	arr[0]=lev1ID
	arr[1],_=createNewBaseStyle(0,lev2)
	arr[2],_=createNewlastTriggerStimulsID(0,lev3,PhraseID,ToneID,MoodID)
	return arr
}
////////////////////////////////////////////////////
